/*
 * out_yasapi.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <initguid.h>
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
//#define YASAPI_NULL

///////////////////////////////////////////////////////////////////////////////
int getwrittentime();
int getoutputtime();

///////////////////////////////////////////////////////////////////////////////
static int active;
static int start_t;
static int last_pause=0;
static volatile int writtentime, w_offset;

static Options options;
static Plugin plugin;
static Player player;

///////////////////////////////////////////////////////////////////////////////
#if 0 // {
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call,
    LPVOID lpReserved)
{
  return TRUE;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
void config(HWND hwnd)
{
#if ! defined (YASAPI_NULL) // {
  ConfigDialog(hwnd,&options);
#endif // }
}

void about(HWND hwnd)
{
#if ! defined (YASAPI_NULL) // {
  AboutDialog(hwnd);
#endif // }
}

void init()
{
  if (OptionsCreate(&options)<0)
    goto create;

  OptionsLoad(&options);

#if defined (YASAPI_DEBUG) // {
  if (options.debug)
    ConsoleOpen();
#endif // }

  DPUTS(options.debug,__FUNCTION__ "\n");
#if defined (_M_IX86_FP) // {
  DPRINTF(options.debug,"  architecture: %d\n",_M_IX86_FP);
#endif // }
  DPRINTF(options.debug,"  options: \"%s\"\n",options.path);
  DPRINTF(options.debug,"    share mode: %s\n",
      YASAPI_SHAREMODE_LABEL(options.eShareMode));
  DPRINTF(options.debug,"    startegy: %s\n",options.pull?"pull":"push");
  DPRINTF(options.debug,"    ring size: %.1f\n",options.qRingSize);
  DPRINTF(options.debug,"    ring filled: %.1f\n",options.qRingFill);
  DPRINTF(options.debug,"    shared size: %.1f\n",options.qShareSize);

#if ! defined (YASAPI_NULL) // {
  if (PluginCreate(&plugin,&options)<0) {
    DMESSAGE(options.debug,"creating plugin");
    goto plugin;
  }
#endif // }

  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

  return;
// cleanup:
#if ! defined (YASAPI_NULL) // {
  PluginDestroy(&plugin);
plugin:
#endif // }
  OptionsDestroy(&options);
create:
  return;
}

void quit()
{
  DPUTS(options.debug,__FUNCTION__ "\n");
#if ! defined (YASAPI_NULL) // {
  PluginDestroy(&plugin);
#endif // }
  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

#if defined (YASAPI_DEBUG) // {
  if (options.debug) {
    Sleep(5000);
    ConsoleClose();
  }
#endif // }
  OptionsDestroy(&options);
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms,
    int prebufferms)
{
  DPUTS(options.debug,__FUNCTION__ "\n");

#if ! defined (YASAPI_NULL) // {
  if (PlayerCreate(&player,&plugin,samplerate,numchannels,bitspersamp)<0) {
    DMESSAGE(options.debug,"creating player");
    goto player;
  }
#else // } {
  WaveFormatExtensibleSetup(&player,samplerate,numchannels,bitspersamp);
#endif // }

  start_t=timeGetTime();
  w_offset = writtentime = 0;
  active=1;
  last_pause=0;

  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");

  return 1;
// cleanup:
#if ! defined (YASAPI_NULL) // {
  PlayerDestroy(&player);
player:
#endif // }
  return -1;
}

void close()
{
  DPUTS(options.debug,__FUNCTION__ "\n");
#if ! defined (YASAPI_NULL) // {
  PlayerDestroy(&player);
#endif // }
  DPUTS(options.debug,"  " __FUNCTION__ " successful\n");
}

int write(char *buf, int len)
{
//DPRINTF(options.debug,__FUNCTION__ ": %d\n",len);
#if ! defined (YASAPI_NULL) // {
  len=PlayerWrite(&player,buf,len);
#endif // }
  writtentime += len;

  return 0;
}

int canwrite()
{
  WAVEFORMATEX *pwfx=&player.wfxx.Format;
  SIZE_T size;

  if (last_pause)
    return 0;

#if ! defined (YASAPI_NULL) // {
  size=PlayerAvailable(&player);
#else // } {
  size=65536;
#endif // }

  if (getwrittentime()<getoutputtime()
      +MulDiv(size,1000,pwfx->nSamplesPerSec*pwfx->nBlockAlign))
    return size;

  return 0;
}

int isplaying()
{
  return 0;
}

int pause(int pause)
{
  int t=last_pause;

  if (last_pause!=pause) {
#if ! defined (YASAPI_NULL) // {
    PlayerPause(&player,pause);
#endif // }

    if (pause) {
      w_offset+=timeGetTime()-start_t;
      writtentime=0;
    }
    else
      start_t=timeGetTime();

    last_pause=pause;
  }

  return t;
}

void setvolume(int volume)
{
  DPUTS(options.debug,__FUNCTION__ "\n");

  if (-666!=volume) {
    plugin.volume=(double)volume/255;
    DPRINTF(options.debug,"  volume: %.2f\n",plugin.volume);
  }
}

void setpan(int pan)
{
}

void flush(int t)
{
  DPUTS(options.debug,__FUNCTION__ "\n");

#if ! defined (YASAPI_NULL) // {
  PlayerReset(&player);
#endif // }
  w_offset=t;
  start_t=timeGetTime();
  writtentime=0;
}
  
int getoutputtime()
{
  if (last_pause)
    return w_offset;

  return timeGetTime()-start_t + w_offset;
}

int getwrittentime()
{
  WAVEFORMATEX *pwfx=&player.wfxx.Format;
  int t=pwfx->nSamplesPerSec*pwfx->nChannels,l;
  int ms=writtentime;

  if (t) {
    l=ms%t;
    ms /= t;
    ms *= 1000;
    ms += (l*1000)/t;

    ms/=player.wBytesPerSample;

    return ms + w_offset;
  }
  else
    return ms;
}

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
  static Out_Module out;

  if (0==out.version) {
    out.version
      // module version (OUT_VER)
      =OUT_VER;
    out.description
      // description of module, with version string
      =PI_LABEL;
    out.id
      // module id. each input module gets its own.
      // non-nullsoft modules should be >= 65536. 
      =65;
    out.hMainWindow
      // winamp's main window (filled in by winamp)
      =NULL;
    out.hDllInstance
      // DLL instance handle (filled in by winamp)
      =NULL;
    out.Config
      // configuration dialog 
      =config;
    out.About
      // about dialog
      =about;
    out.Init
      // called when loaded
      =init;
    out.Quit
      // called when unloaded
      =quit;
    out.Open
      // returns >=0 on success, <0 on failure
      //
      // NOTENOTENOTE: bufferlenms and prebufferms are ignored in most
      // if not all output plug-ins. 
      //    ... so don't expect the max latency returned to be what you
      // asked for.
      // returns max latency in ms (0 for diskwriters, etc)
      // bufferlenms and prebufferms must be in ms. 0 to use defaults. 
      // prebufferms must be <= bufferlenms
      // pass bufferlenms==-666 to tell the output plugin that it's clock
      // is going to be used to sync video
      // out_ds turns off silence-eating when -666 is passed
      =open;
    out.Close
      // close the ol' output device.
      =close;
    out.Write
      // 0 on success. Len == bytes to write (<= 8192 always). buf is
      //   straight audio data. 
      // 1 returns not able to write (yet). Non-blocking, always.
      =write;
    out.CanWrite
      // returns number of bytes possible to write at a given time. 
      // Never will decrease unless you call Write (or Close, heh)
      =canwrite;
    out.IsPlaying
      // non0 if output is still going or if data in buffers waiting to be
      // written (i.e. closing while IsPlaying() returns 1 would truncate
      // the song
      =isplaying;
    out.Pause
      // returns previous pause state
      =pause;
    out.SetVolume
      // volume is 0-255
      =setvolume;
    out.SetPan
      // pan is -128 to 128
      =setpan;
    out.Flush
      // flushes buffers and restarts output at time t (in ms) 
      // (used for seeking)
      =flush;
    out.GetOutputTime
      // returns played time in MS
      =getoutputtime;
    out.GetWrittenTime
      // returns time written in MS (used for synching up vis stuff)
      =getwrittentime;
  }

  return &out;
}
