/*
 * yasapi_options.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

#define YASAPI_OPTION_FILE "Plugins\\" YASAPI_NAME ".ini"

#define YASAPI_APP_PLUGIN         "plugin"
#define YASAPI_KEY_VOLUME         "volume"
#define YASAPI_KEY_DEBUG          "debug"

#define YASAPI_APP_WASAPI         "wasapi"
#define YASAPI_KEY_SHARE_MODE     "share_mode"
#define YASAPI_KEY_PULL           "pull"
#define YASAPI_KEY_RING_SIZE      "ring_size"
#define YASAPI_KEY_RING_FILL      "ring_fill"
#define YASAPI_KEY_SHARE_SIZE     "share_size"

int OptionsCreate(Options *pOptions)
{
  char *dir;
  size_t len1,len2;
  char *wp;

  ZeroMemory(
    pOptions,           // [in]  PVOID Destination,
    sizeof *pOptions    // [in]  SIZE_T Length
  );

  dir=(char *)SendMessageW(winampGetOutModule()->hMainWindow,
      WM_WA_IPC,0,IPC_GETINIDIRECTORY);

  if (NULL==dir)
    goto dir;

  len1=strlen(dir);
  len2=strlen(YASAPI_OPTION_FILE);

  pOptions->path=(LPSTR)GlobalAlloc(
    GMEM_FIXED,         // _In_  UINT uFlags,
    (len1+1)+(len2+1)   // _In_  SIZE_T dwBytes
  );

  if (NULL==pOptions->path)
    goto path;
  
  wp=pOptions->path;

  memcpy(wp,dir,len1);
  wp+=len1;
  *wp++='\\';

  memcpy(wp,YASAPI_OPTION_FILE,len2);
  wp+=len2;
  *wp++=0;

  return 0;
// cleanup:
  GlobalFree(pOptions->path);
path:
dir:
  return -1;
}

void OptionsDestroy(Options *pOptions)
{
  GlobalFree(pOptions->path);
}

void OptionsLoad(Options *pOptions)
{
  char set[32];
  char get[32];

#if defined (YASAPI_DEBUG) // {
  // get the debug flag.
  pOptions->debug=GetPrivateProfileIntA(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_DEBUG,             // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->debug,   // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );
#endif // }

  // get the share mode.
  pOptions->eShareMode=GetPrivateProfileIntA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_MODE,        // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->eShareMode,
                                  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the pull flag.
  pOptions->pull=GetPrivateProfileIntA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PULL,              // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->pull,    // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the volume flag.
  pOptions->volume=GetPrivateProfileIntA(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_VOLUME,            // _In_  LPCTSTR lpKeyName,
    GetDefaultOptions()->volume,  // _In_  INT nDefault,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // get the ring size.
  sprintf(set,"%f",GetDefaultOptions()->qRingSize);

  GetPrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_RING_SIZE,         // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    sizeof get,                   // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qRingSize=atof(get);

  // get the ring fill.
  sprintf(set,"%f",GetDefaultOptions()->qRingFill);

  GetPrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_RING_FILL,         // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    sizeof get,                   // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qRingFill=atof(get);

  // get the share size.
  sprintf(set,"%f",GetDefaultOptions()->qShareSize);

  GetPrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_   LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_SIZE,        // _In_   LPCTSTR lpKeyName,
    set,                          // _In_   LPCTSTR lpDefault,
    get,                          // _Out_  LPTSTR lpReturnedString,
    sizeof get,                   // _In_   DWORD nSize,
    pOptions->path                // _In_   LPCTSTR lpFileName
  );

  pOptions->qShareSize=atof(get);
}

#if defined (YASAPI_DEBUG) // {
void OptionsSave(Options *pOptions, int debug)
#else // } {
void OptionsSave(Options *pOptions)
#endif // }
{
  char buf[32];

  // write the share size.
  sprintf(buf,"%f",pOptions->qShareSize);

  WritePrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_SIZE,        // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the ring fill.
  sprintf(buf,"%f",pOptions->qRingFill);

  WritePrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_RING_FILL,         // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the ring size.
  sprintf(buf,"%f",pOptions->qRingSize);

  WritePrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_RING_SIZE,         // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the volume flag.
  sprintf(buf,"%d",pOptions->volume);

  WritePrivateProfileStringA(
    YASAPI_APP_PLUGIN,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_VOLUME,            // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the pull flag.
  sprintf(buf,"%d",pOptions->pull);

  WritePrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_PULL,              // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

  // write the share mode.
  sprintf(buf,"%d",pOptions->eShareMode);

  WritePrivateProfileStringA(
    YASAPI_APP_WASAPI,            // _In_  LPCTSTR lpAppName,
    YASAPI_KEY_SHARE_MODE,        // _In_  LPCTSTR lpKeyName,
    buf,                          // _In_  LPCTSTR lpString,
    pOptions->path                // _In_  LPCTSTR lpFileName
  );

#if defined (YASAPI_DEBUG) // {
  if (pOptions->debug!=debug) {
    sprintf(buf,"%d",pOptions->debug);

    WritePrivateProfileStringA(
      YASAPI_APP_PLUGIN,          // _In_  LPCTSTR lpAppName,
      YASAPI_KEY_DEBUG,           // _In_  LPCTSTR lpKeyName,
      buf,                        // _In_  LPCTSTR lpString,
      pOptions->path              // _In_  LPCTSTR lpFileName
    );

    if (0==pOptions->debug)
      ConsoleClose();
    else
      ConsoleOpen();
  }
#endif // }
}

#if 0 // {
void OptionsCopy(Options *lhs, const Options *rhs)
{
  char *path;

  path=lhs->path;
  *lhs=*rhs;
  lhs->path=path;
}
#endif // }

const Options *GetDefaultOptions(void)
{
  static int initialized=0;
  static Options options;

  if (0==initialized) {
#if defined (YASAPI_DEBUG) // {
    options.debug=TRUE;
#endif // }
    options.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    options.pull=FALSE;
    options.volume=FALSE;
    options.qRingSize=5.0;
    options.qRingFill=0.0;
    options.qShareSize=0.0;
    initialized=1;
  }

  return &options;
}
